/* Name: main.c
 * Author: James Fletcher
 * Copyright: (c) James Fletcher
 * License: This work may be reproduced, modified, distributed, performed, and displayed for any purpose, but must acknowledge "fab-synth". Copyright is retained and must be preserved. The work is provided as is; no warranty is provided, and users accept all liability.
 */


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "sine-wave.h"

// IO pin usage
#define CV_SYNC_IN      _BV(PB2)

#define AUDIO_OUT_PIN   _BV(PA7) // Audio PWM out on OC0B (PA7)

#define SPI_DI          _BV(PA6)
#define SPI_DO          _BV(PA5)
#define SPI_SCK         _BV(PA4)

#define LED_PIN         _BV(PA3)

#define POT3            _BV(PA2)
#define POT2            _BV(PA1)
#define POT1            _BV(PA0)


/*
 * SPI frame format for audio and cv data
 * Total of 20 bytes timeslots are:
 *   sync, cv-hi, cv-mid, cv-lo, aud0, ... aud15
 * Sync timeslot is empty, it just contains a falling edge on the sync line signaling the start of a new frame
 */
#define FRAME_SYNC_SLOT     0
#define FRAME_CV0_SLOT      1
#define FRAME_CV1_SLOT      2
#define FRAME_CV2_SLOT      3
#define FRAME_AUDIO_START_SLOT   4
// We have 3 frame buffers:
// 1 - currently receiving data
// 2 - currently DSP proccessing data
// 3 - currently sending to the next node
uint8_t rxFrame[20];
uint8_t processingFrame[20];
uint8_t txFrame[21]; // we que up byte beyond end of buffer but never send, hence 1 more byte here
uint8_t rxFrameSlotIndex = 0; // index of time slot in frame
volatile uint8_t bufferReady = 0; // semaphore to main code that a new buffer is ready to process

// CV data
// cv0: bit 7 - trigger
// cv0: bit 6 - gate
// cv1: unused
// cv2: control voltage value
#define CV0_TRIG 0x20
#define CV0_GATE 0x40

// Values of pots, updated live
uint16_t pots[3];
#define POT_TUNE        0
#define POT_WAVESELECT  1
#define POT_RANGE       2

// Current waveform type
#define WAVE_SQUARE     0
#define WAVE_TRI        1
#define WAVE_SAW        2
#define WAVE_SINE       3
volatile uint8_t waveType = WAVE_SQUARE;

// Audio PWM
uint8_t audioPWMSlot = 0;

// Waveform step, used to build oscillator waveforms.
// waveIndexL is an 8:8 fixed point, with int part 0 -> 255 being one complete waveform cycle.
uint16_t waveIndexL = 0;
uint16_t waveStepRate = 2<<8; // 8:8 step rate through waveforms
uint16_t waveStepRateNew = 2<<8; // 8:8 step rate through waveforms - this is swapped into waveStepRate near a zero crossing



void cycleFrameBuffers(void);



// Update PWM output - old code
// OCR0B is double buffered and only updated as counter crosses bottom, so no need to update here
ISR(TIM0_COMPB_vect)
{
}


// Timer 1 compare/overflow - update audio PWM duty cycle on Timer 0 with next audio sample
ISR(TIM1_COMPA_vect)
{
    // Update waveform output
    audioPWMSlot++;
    audioPWMSlot &= 0x0f; // range limit for case when no clock module present
    OCR0B = txFrame[FRAME_AUDIO_START_SLOT+audioPWMSlot];
    
}


// USI counter overflow indicates SPI byte rx
ISR(USI_OVF_vect)
{
    // Clear irq flag
	USISR = _BV(USIOIF);
    
    // Move rx byte into buffer
    rxFrame[rxFrameSlotIndex] = USIDR;
    
    // Update index
    rxFrameSlotIndex++;
    
    // Send next tx byte
    USIDR = txFrame[rxFrameSlotIndex]; // may be byte 21 hence larger tx buffer

    if(rxFrameSlotIndex >= 20)
    {
        cycleFrameBuffers(); // Move frame somewhere for processing
        
        // Signal new frame arrived to DSP code
        bufferReady = 1;
        
        // debug
        //if(rxFrame[3] != 0)
        //    PORTA &= ~LED_PIN; // Some key, LED on.
        //else
        //    PORTA |= LED_PIN; // LED off
        rxFrameSlotIndex = 1; // byte 0 (sync) never sent
    }
}

// INT0 pin IRQ
// SYNC signal from clock unit
// Setup to generate IRQ on falling edge of sync pin
ISR(EXT_INT0_vect)
{
    // Re-sync audio smaple rate clock in Timer1
    // TCNT1 irq should be due to fire any cycle now, so reset counter as if irq had fired and put 1st byte on tx frame into DAC
    TCNT1 = 0;
    audioPWMSlot = 0;
    OCR0B = txFrame[FRAME_AUDIO_START_SLOT+audioPWMSlot];
    
    
    // Output 1st byte of our frame ready for transfer session
    USIDR = txFrame[1];

    // Reset to start of frame
    rxFrameSlotIndex = 1; // byte 0 (sync) never sent
    // Reset SPI rx to start of a new byte
    USISR = 0b11100000; // clear flags, zero counter
}


// Called each time we recieve a new frame from SPI
void cycleFrameBuffers(void)
{
    int i;
    
    // Move all frame buffers around
    // No need to move byte 0 (Sync) as it's empty
    for(i=1; i<20; i++)
    {
        txFrame[i] = processingFrame[i];
        processingFrame[i] = rxFrame[i];
    }
}


// Zero out the audio signal as no note at the moment
void zeroOscSignal(void)
{
    uint8_t i;
    
    for(i=0; i<16; i++)
        processingFrame[FRAME_AUDIO_START_SLOT+i] = 0;
}


// Place next batch of 16 audio samples into the output stream
// 1) waveIndex 0 -> 256 is one complete waveform cycle, so we scale this based on requested frequency
// 2) force a frequency swap near a zero crossing point by only allowing frequency change when waveIndex near 0
void fillInOscSignal(void)
{
    uint8_t i;
    uint8_t b=0;
    uint8_t waveIndex;
    
    for(i=0; i<16; i++)
    {
        waveIndex = waveIndexL >> 8; // get integer part of current position in waveform
        
        if(waveType == WAVE_SAW)
            b = waveIndex;
        else if (waveType == WAVE_SQUARE)
        {
            if(waveIndex < 128)
                b = 0;
            else
                b = 250;
        }
        else if(waveType == WAVE_TRI)
        {
            if(waveIndex < 128)
                b = waveIndex*2;
            else
                b = (255 - waveIndex)*2;
        }
        else if(waveType == WAVE_SINE)
            b = pgm_read_byte(&sineWave[waveIndex]);
        
        processingFrame[FRAME_AUDIO_START_SLOT+i] = b; // place audio sample in output stream
        
        // Update position in waveform (16 bit number so wraps naturally at full scale)
        waveIndexL += waveStepRate;
        
        // Check for a near zero crossing
        if(waveIndexL <= waveStepRate)
        {
            if(waveStepRate != waveStepRateNew)
            {
                waveStepRate = waveStepRateNew;
                waveIndexL = 0;
            }
        }
    }
}


// When pots are used as selector switches this fn converts the ADC value (0->1024) into a selector value from the range.
// It applies some hysteresis based on current selected value
// Should pre-calc these range values and fill in a table
uint8_t potSelector(uint8_t range, uint8_t currentSelected, uint16_t potValue)
{
    uint16_t sizeOfSelectorBands = 1024 / range;
    uint8_t selected;
    
    selected = (potValue*range)/1024;
    
    if(selected == currentSelected+1)
    {
        // moved one band up, apply some hysteresis
        uint16_t remainder = (potValue*range)%1024;
        if(remainder < 100) // 10%
            selected = currentSelected;
    }
    else if(selected == currentSelected-1) // don't worry about negative underflow on our un-signed type
    {
        // moved one band down, apply some hysteresis
        uint16_t remainder = (potValue*range)%1024;
        if(remainder > 924) // 10%
            selected = currentSelected;
    }
    // Otherwise band is either the same or much furtehr away, so no need to calc hysteresis
    
    return selected;
}



int main(void)
{
    uint8_t i;
    uint8_t current_ADC_Chan = 0;

    
    // Zero frame buffers
    for(i=0; i<20; i++)
        rxFrame[i] = processingFrame[i] = txFrame[i] = 0;
    
    // Zero pots
    for(i=0; i<3; i++)
        pots[i] = 0;
    
    // Setup outputs
    DDRA = LED_PIN | AUDIO_OUT_PIN | SPI_DO;
    //DDRB = CV_SYNC_IN;
    
    // Enable pull ups on SPI input pins?
    PORTA = SPI_DI | SPI_SCK;
    
    // Turn LED off
    PORTA |= LED_PIN; //&= ~LED_PIN;
    
    
    // TCNT0 Generates our fast 78KHz PWM signal for our DAC
    // Setup TCNT0 for PWM Fast mode
    // Timer cycles at 78125Hz
    TCCR0A = 0b00100011; // Clear OC0B on match, Fast PWM (mode 3)
    TCCR0B = 0b00000001; // Fast PWM, No prescale
    TCNT0 = 0;
    OCR0B = 128+64;
    //TIMSK0 = 0b00000100; // Enable OC0B match irq - no longer update OC0B here
    
    // TCNT1 Generates PWM duty cycle update events, ie. times out at the audio sample frequency of 31KHz
    // I had wanted to use SPI rx irq for this, but now we have our cv data in the audio stream there is no longer a 1:1 between SPI bytes and audio sample points
    // SPI bytes turn up at 39062.5Hz.  16 out of every 20 are audio samples giveing 31250Hz audio sample rate
    TCCR1A = 0b00000000; // No OC1x functions, CTC (Mode 4)
    TCCR1B = 0b00001001; // CTC, no pre-scaler
    TCNT1 = 0;
    OCR1A = 700; //F_CPU/31250; // 640 - gives 31250Hz
    TIMSK1 = 0b00000010; // enable compare A irq

    // Enable SPI
    USICR = 0b01011000; // irq on overflow, 3 wire mode, ext clock pos edge
    USISR = 0b11100000; // clear flags, zero counter
    
    // Enable INT0 (sync pin), falling edge generates irq
    MCUCR = (MCUCR & 0xfc) | 0x02;
    GIMSK |= _BV(INT0);
    
    sei(); // Global IRQ enable
    
    for(;;)
    {
        // Read the 3 pots and store the values
        // Force long ADC conversion as we're swapping channels.  Not sure it's really required for single ended inputs
        //ADCSRA = 0;
        // Set ADC channel
        ADMUX = 0b00000000 | (current_ADC_Chan& 0x1f); // Ref=VCC, channel = current
        ADCSRA = 0b10000100; // ADC enable, no irqs, prescale / 16 (need to slow ADC slightly as my pots are 50K and take a while to charge the SH unit)
        ADCSRA |= (1 << ADSC); // Initiate conversion
        while (ADCSRA & (1 << ADSC)) // Wait for it to complete
            ;
        pots[current_ADC_Chan] = 1024 - ADC; // Pots are soldered in back to front, so invert the range
        current_ADC_Chan++;
        if(current_ADC_Chan > 2) current_ADC_Chan = 0;
        
#if 1
        waveType = potSelector(4, waveType, pots[POT_WAVESELECT]);
#else
        if(pots[POT_WAVESELECT] < 256)
            waveType = WAVE_SQUARE;
        else if(pots[POT_WAVESELECT] < 512)
            waveType = WAVE_TRI;
        else if(pots[POT_WAVESELECT] < 768)
            waveType = WAVE_SAW;
        else
            waveType = WAVE_SINE;
#endif
        
        if((waveType == WAVE_SAW ) || (waveType == WAVE_SQUARE))
            PORTA &= ~LED_PIN; // Some key, LED on.
        else
            PORTA |= LED_PIN; // LED off
        
        // Check for new procesisng to be done
        if(bufferReady > 0)
        {
            bufferReady = 0;
            
            if((processingFrame[FRAME_CV0_SLOT] & CV0_GATE) != 0)
            {
                // work out freq
                waveStepRateNew = ((uint16_t)processingFrame[FRAME_CV2_SLOT]) << 4;
            }
            // else leave freq alone
            
            //if((processingFrame[FRAME_CV0_SLOT] & CV0_GATE) == 0)
            //    zeroOscSignal();
            //else
                fillInOscSignal();
            
        }

    }
    
    return 0;
}
